"""
Core trace list implementation for pymseed

"""

from __future__ import annotations

from collections.abc import Sequence
from time import time
from typing import Any, Callable, Optional

from .clib import cdata_to_string, clibmseed, ffi
from .definitions import DataEncoding, SubSecond, TimeFormat
from .exceptions import MiniSEEDError
from .msrecord import MS3Record
from .util import encoding_sizetype, nstime2timestr


class MS3RecordPtr:
    """Wrapper around CFFI MS3RecordPtr structure"""

    def __init__(self, cffi_ptr: Any) -> None:
        self._ptr = cffi_ptr

    def __repr__(self) -> str:
        return (
            f"MS3RecordPtr(sourceid: {self.record.sourceid}\n"
            f"             filename: {cdata_to_string(self._ptr.filename)}\n"
            f"           fileoffset: {self._ptr.fileoffset}\n"
            f"            bufferptr: {self._ptr.bufferptr}\n"
            f"               reclen: {self._ptr.msr.reclen}\n"
            f"            starttime: {self.record.starttime_str(timeformat=TimeFormat.ISOMONTHDAY_Z)}\n"
            f"              endtime: {self.record.endtime_str(timeformat=TimeFormat.ISOMONTHDAY_Z)}\n"
            ")"
        )

    def __str__(self) -> str:
        return (
            f"{self.record.sourceid}, "
            f"{cdata_to_string(self._ptr.filename)}, "
            f"byte offset: {self._ptr.fileoffset}"
        )

    @property
    def record(self) -> MS3Record:
        """Return a constructed MS3Record"""
        if not hasattr(self, "_msrecord"):
            self._msrecord = MS3Record(recordptr=self._ptr.msr)
        return self._msrecord

    @property
    def filename(self) -> Optional[str]:
        """Return filename as string"""
        result = cdata_to_string(self._ptr.filename)
        if result is None:
            return None
        return result

    @property
    def fileoffset(self) -> int:
        """Return file offset"""
        return self._ptr.fileoffset

    @property
    def endtime(self) -> int:
        """Return end time"""
        return self._ptr.endtime

    @property
    def dataoffset(self) -> int:
        """Return data offset"""
        return self._ptr.dataoffset


class MS3RecordList:
    """Wrapper around CFFI MS3RecordList structure

    This class supports list-like access to the record pointers:
    - len(record_list) returns the number of records
    - record_list[i] returns the i-th record pointer
    - record_list[start:end] returns a slice of record pointers
    - for record_ptr in record_list: iterates over all record pointers
    """

    def __init__(self, cffi_ptr: Any) -> None:
        self._list = cffi_ptr

    def __repr__(self) -> str:
        def indent_repr(thing):
            """Add two-space indentation to each line of repr(thing)"""
            return "\n".join("  " + line for line in repr(thing).split("\n"))

        # Create list of formatted strings
        if len(self) <= 5:
            formatted_lines = [indent_repr(recptr) for recptr in self]
        else:
            formatted_lines = [
                indent_repr(self[0]),
                indent_repr(self[1]),
                f"  ... {len(self) - 4} more",
                indent_repr(self[-2]),
                indent_repr(self[-1]),
            ]

        newline = "\n"
        return f"MS3RecordList(recordcnt: {len(self)}\n{newline.join(formatted_lines)}\n)"

    def __str__(self) -> str:
        def indent_str(thing):
            """Add two-space indentation to each line of str(thing)"""
            return "\n".join("  " + line for line in str(thing).split("\n"))

        # Create list of formatted strings
        if len(self) <= 5:
            formatted_lines = [indent_str(recptr) for recptr in self]
        else:
            formatted_lines = [
                indent_str(self[0]),
                indent_str(self[1]),
                f"  ... {len(self) - 4} more",
                indent_str(self[-2]),
                indent_str(self[-1]),
            ]

        newline = "\n"
        return f"Record list with {len(self)} records\n{newline.join(formatted_lines)}"

    def __len__(self) -> int:
        """Return number of records"""
        return self._list.recordcnt

    def __iter__(self) -> Any:
        """Return iterator over record pointers"""
        current_record = self._list.first
        while current_record != ffi.NULL:
            yield MS3RecordPtr(current_record)
            current_record = current_record.next

    def __getitem__(self, key: int | slice) -> Any:
        """Enable indexing and slicing access to record pointers"""
        if isinstance(key, slice):
            # Handle slice objects (e.g., record_list[1:3], record_list[::2])
            record_list = list(self)
            return record_list[key]
        elif isinstance(key, int):
            # Handle single integer index
            length = len(self)
            if length == 0:
                raise IndexError("list index out of range")

            # Handle negative indices
            if key < 0:
                key += length

            # Check bounds
            if key < 0 or key >= length:
                raise IndexError("list index out of range")

            # Find and return the record at the specified index
            for i, record in enumerate(self):
                if i == key:
                    return record

            # This shouldn't happen if our logic is correct
            raise IndexError("list index out of range")
        else:
            raise TypeError("indices must be integers or slices")

    @property
    def recordcnt(self) -> int:
        """Return record count"""
        return self._list.recordcnt

    def records(self) -> Any:
        """Return the records via a generator iterator"""
        current_record = self._list.first
        while current_record != ffi.NULL:
            yield MS3RecordPtr(current_record)
            current_record = current_record.next


class MS3TraceSeg:
    """Wrapper around CFFI MS3TraceSeg structure"""

    def __init__(
        self, cffi_ptr: Any, parent_id_ptr: Any = None, parent_tracelist: Any = None
    ) -> None:
        self._seg = cffi_ptr
        self._parent_id = parent_id_ptr  # Reference to parent MS3TraceID
        self._parent_tracelist = parent_tracelist  # Reference to parent MS3TraceList

    def __repr__(self) -> str:
        sample_preview = "[]"
        if self.numsamples > 0:
            if len(self.datasamples) > 5:
                # Create array representation with ellipsis inside: [1,2,3,4,5,...]
                first_samples = ", ".join(str(sample) for sample in list(self.datasamples[:5]))
                sample_preview = f"[{first_samples}, ...]"
            else:
                sample_preview = str(list(self.datasamples))

        return (
            f"MS3TraceSeg(start: {self.starttime_str(timeformat=TimeFormat.ISOMONTHDAY_DOY_Z)}\n"
            f"              end: {self.endtime_str(timeformat=TimeFormat.ISOMONTHDAY_DOY_Z)}\n"
            f"         samprate: {self.samprate}\n"
            f"        samplecnt: {self.samplecnt}\n"
            f"      datasamples: {sample_preview}\n"
            f"         datasize: {self.datasize}\n"
            f"       numsamples: {self.numsamples}\n"
            f"       sampletype: {self.sampletype}\n"
            f"       recordlist: {'Record list of ' + str(len(self.recordlist)) + ' records' if self.recordlist else 'None'}\n"
            ")"
        )

    def __str__(self) -> str:
        return (
            f"start: {self.starttime_str(timeformat=TimeFormat.ISOMONTHDAY_DOY_Z)}, "
            f"end: {self.endtime_str(timeformat=TimeFormat.ISOMONTHDAY_DOY_Z)}, "
            f"samprate: {self.samprate}, "
            f"samples: {self.samplecnt} "
        )

    @property
    def starttime(self) -> int:
        """Return start time as nanoseconds since Unix/POSIX epoch"""
        return self._seg.starttime

    @property
    def starttime_seconds(self) -> float:
        """Return start time as seconds since Unix/POSIX epoch"""
        return self._seg.starttime / clibmseed.NSTMODULUS

    def starttime_str(
        self,
        timeformat: TimeFormat = TimeFormat.ISOMONTHDAY_Z,
        subsecond: SubSecond = SubSecond.NANO_MICRO_NONE,
    ) -> Optional[str]:
        """Return start time as formatted string"""
        result = nstime2timestr(self._seg.starttime, timeformat, subsecond)
        if result is None:
            return None
        return result

    @property
    def endtime(self) -> int:
        """Return end time as nanoseconds since Unix/POSIX epoch"""
        return self._seg.endtime

    @property
    def endtime_seconds(self) -> float:
        """Return end time as seconds since Unix/POSIX epoch"""
        return self._seg.endtime / clibmseed.NSTMODULUS

    def endtime_str(
        self,
        timeformat: TimeFormat = TimeFormat.ISOMONTHDAY_Z,
        subsecond: SubSecond = SubSecond.NANO_MICRO_NONE,
    ) -> Optional[str]:
        """Return end time as formatted string"""
        result = nstime2timestr(self._seg.endtime, timeformat, subsecond)
        if result is None:
            return None
        return result

    @property
    def samprate(self) -> float:
        """Return sample rate"""
        return self._seg.samprate

    @property
    def samplecnt(self) -> int:
        """Return sample count"""
        return self._seg.samplecnt

    @property
    def recordlist(self) -> Optional[MS3RecordList]:
        """Return the record list structure"""
        if self._seg.recordlist:
            return MS3RecordList(self._seg.recordlist)
        else:
            return None

    @property
    def datasamples(self) -> memoryview:
        """Return data samples as a memoryview (no copy)

        A view of the data samples in an internal buffer owned by this MS3Record
        instance is returned.  If the data are needed beyond the lifetime of this
        instance, a copy must be made.

        The returned view can be used directly with slicing and indexing
        from `0` to `MS3TraceSeg.numsamples - 1`.

        The view can efficiently be copied to a _python list_ using:

            data_samples = MS3TraceSeg.datasamples[:]
        """
        if self._seg.numsamples <= 0:
            return memoryview(b"")  # Empty memoryview

        sampletype = self.sampletype

        if sampletype == "i":
            ptr = ffi.cast("int32_t *", self._seg.datasamples)
            buffer = ffi.buffer(ptr, self._seg.numsamples * ffi.sizeof("int32_t"))
            return memoryview(buffer).cast("i")
        elif sampletype == "f":
            ptr = ffi.cast("float *", self._seg.datasamples)
            buffer = ffi.buffer(ptr, self._seg.numsamples * ffi.sizeof("float"))
            return memoryview(buffer).cast("f")
        elif sampletype == "d":
            ptr = ffi.cast("double *", self._seg.datasamples)
            buffer = ffi.buffer(ptr, self._seg.numsamples * ffi.sizeof("double"))
            return memoryview(buffer).cast("d")
        elif sampletype == "t":
            ptr = ffi.cast("char *", self._seg.datasamples)
            buffer = ffi.buffer(ptr, self._seg.numsamples)
            return memoryview(buffer).cast("B")
        else:
            raise ValueError(f"Unknown sample type: {sampletype}")

    @property
    def sampletype(self) -> Optional[str]:
        """Return sample type code if available, otherwise None"""
        if self._seg.sampletype:
            return str(self._seg.sampletype.decode("ascii"))
        else:
            return None

    @property
    def numsamples(self) -> int:
        """Return number of samples"""
        return self._seg.numsamples

    @property
    def datasize(self) -> int:
        """Return data size in bytes"""
        return self._seg.datasize

    @property
    def sample_size_type(self) -> tuple[int, str]:
        """Return data sample size and type code from first record in list

        NOTE: This is a guesstimate based on the first record in the record list.
        It is not guaranteed to be correct for any other records in the list.
        """
        if self._seg.recordlist is None:
            raise ValueError("No record list available to determine sample size and type")

        # Get the first record
        first_record_ptr = self._seg.recordlist.first

        if first_record_ptr is None:
            raise ValueError("No records in record list")

        return encoding_sizetype(first_record_ptr.msr.encoding)

    @property
    def np_datasamples(self) -> Any:
        """Return data samples as a numpy array (no copy)

        A view of the data samples in an internal buffer owned by this MS3TraceSeg
        instance is returned. If the data are needed beyond the lifetime of this
        instance, a copy must be made.
        """
        try:
            import numpy as np
        except ImportError:
            raise ImportError(
                "numpy is not installed. Install numpy or this package with [numpy] optional dependency"
            ) from None

        if self._seg.numsamples <= 0:
            return np.array([])  # Empty array

        sampletype = self.sampletype

        # Translate libmseed sample type to numpy type
        nptype = {
            "i": np.int32,
            "f": np.float32,
            "d": np.float64,
            "t": "S1",  # 1-byte strings for text data
        }

        if sampletype not in nptype:
            raise ValueError(f"Unknown sample type: {sampletype}")

        # Create numpy array view from CFFI buffer
        return np.frombuffer(self.datasamples, dtype=nptype[sampletype])

    def create_numpy_array_from_recordlist(self) -> Any:
        """Return data samples as a numpy array unpacked from the record list

        The numpy array returned is an independent copy of the data samples.
        """
        try:
            import numpy as np
        except ImportError:
            raise ImportError(
                "numpy is not installed. Install numpy or this package with [numpy] optional dependency"
            ) from None

        if self.recordlist is None:
            raise ValueError(
                "Record list required, use record_list=True when populating MS3TraceList"
            )

        if self.samplecnt <= 0:
            return np.array([])  # Empty array

        (_, sample_type) = self.sample_size_type

        # Translate libmseed sample type to numpy type
        nptype = {
            "i": np.int32,
            "f": np.float32,
            "d": np.float64,
            "t": "S1",  # 1-byte strings for text data
        }

        if sample_type not in nptype:
            raise ValueError(f"Unknown sample type: {sample_type}")

        # Create numpy array of the correct type and size
        array = np.empty(self.samplecnt, dtype=nptype[sample_type])

        # Unpack data samples into the array
        self.unpack_recordlist(buffer=array)

        return array

    def unpack_recordlist(self, buffer: Any = None, verbose: int = 0) -> int:
        """Unpack data samples from miniSEED record list into accessible format

        This method decodes data samples from the original miniSEED records that were
        stored when reading with `record_list=True`. It's used for memory-efficient
        workflows where you delay data unpacking until needed.

        Args:
            buffer: Optional destination buffer for unpacked data. Must support the
                buffer protocol (e.g., numpy array, bytearray, memoryview). If provided,
                must be large enough to hold `self.samplecnt` samples of the appropriate
                data type. If None, data is unpacked into internal memory owned by this
                segment instance.
            verbose: Verbosity level for diagnostic output (0=quiet, 1-3=increasing
                detail). Default: 0

        Returns:
            Number of samples successfully unpacked

        Raises:
            ValueError: If no record list is available (requires `record_list=True` when
                reading), if data is already unpacked and a buffer is provided, or if
                the provided buffer doesn't support the buffer protocol
            MiniSEEDError: If unpacking fails due to corrupted or invalid record data

        Note:
            - Requires the segment to have been created with `record_list=True`
            - Can only be called once per segment if using internal memory (buffer=None)
            - If using a provided buffer, the buffer format must match the segment's
              sample type (int32 for "i", float32 for "f", float64 for "d", bytes for "t")
            - For performance, use memoryviews with matching dtype when providing buffers

        Examples:
            Basic workflow illustrating how to use unpack_recordlist():

            >>> from pymseed import MS3TraceList

            Basic unpacking to internal memory:
            >>> traces = MS3TraceList.from_file("examples/example_data.mseed", record_list=True)
            >>> len(traces)
            3
            >>> # Before unpacking, the data samples are not available
            >>> for traceid in traces:
            ...     for segment in traceid:
            ...         assert(segment.datasamples == memoryview(b''))
            ...         assert(segment.numsamples == 0)

            >>> # After unpacking, the data samples are available
            >>> for traceid in traces:
            ...     for segment in traceid:
            ...         count = segment.unpack_recordlist()
            ...         assert(segment.numsamples == segment.samplecnt)
            ...         assert(len(segment.datasamples) == segment.numsamples)

            Advanced example of unpacking data to a numpy array:
            Note: this example is for illustration only.  If numpy arrays are desired
            use the provided create_numpy_array_from_recordlist() or np_datasamples() instead.

            >>> # For doctest conditional skipping, unneeded for real code
            >>> try:
            ...     import numpy as np
            ...     HAS_NUMPY = True
            ... except ImportError:
            ...     HAS_NUMPY = False

            >>> if HAS_NUMPY:
            ...     traces = MS3TraceList.from_file("examples/example_data.mseed", record_list=True)
            ...     for traceid in traces:
            ...         for segment in traceid:
            ...             # Get the sample size and type from the first record in the record list
            ...             (size, sample_type) = segment.sample_size_type
            ...
            ...             if sample_type == "i":
            ...                 # Create a numpy array to hold the unpacked data
            ...                 numpy_array = np.zeros(segment.samplecnt, dtype=np.int32)
            ...
            ...                 # Unpack the data directly into our array
            ...                 count = segment.unpack_recordlist(numpy_array)
            ...
            ...                 # Check that the array is not all zeros
            ...                 assert not np.all(numpy_array == 0), "Numpy array is all zeros (should not happen)"
            ...
            ...             # Other sample types would need different numpy array types

            # Advanced example of unpacking data to an Apache Arrow array using pyarrow
            >>> # For doctest conditional skipping, unneeded for real code
            >>> try:
            ...     import pyarrow as pa
            ...     HAS_PYARROW = True
            ... except ImportError:
            ...     HAS_PYARROW = False

            >>> if HAS_PYARROW:
            ...     traces = MS3TraceList.from_file("examples/example_data.mseed", record_list=True)
            ...     for traceid in traces:
            ...         for segment in traceid:
            ...             # Get the sample size and type from the first record in the record list
            ...             (size, sample_type) = segment.sample_size_type
            ...
            ...             if sample_type == "i":
            ...                 # Create a pyarrow array to hold the unpacked data
            ...                 pyarrow_array = pa.array([0] * segment.samplecnt, type=pa.int32())
            ...
            ...                 # Get the data buffer for direct writing
            ...                 array_bitmap, array_buffer = pyarrow_array.buffers()
            ...
            ...                 # Unpack the data directly into our array buffer
            ...                 count = segment.unpack_recordlist(array_buffer)
            ...
            ...                 # Check that the array has no nulls (all values are valid)
            ...                 assert pyarrow_array.null_count == 0, "Pyarrow array has nulls (should not happen)"
            ...
            ...             # Other sample types would need different pyarrow array types

            The point of the advanced examples is to illustrate how to use the pattern
            to unpack data into a buffer provided by the caller in order to avoid
            copying the data.
        """
        if self.recordlist is None:
            raise ValueError("No record list available to unpack")

        if self.datasamples and buffer is not None:
            raise ValueError("Data samples already unpacked")

        # Handle buffer types that may not be compatible with memoryview
        buffer_ptr = ffi.NULL
        buffer_size = 0
        if buffer is not None:
            try:
                buffer_ptr = ffi.from_buffer(buffer)
                buffer_size = buffer.nbytes
            except (TypeError, AttributeError):
                # Try to get size through len() if nbytes is not available
                try:
                    buffer_ptr = ffi.from_buffer(buffer)
                    buffer_size = (
                        len(buffer) * buffer.itemsize
                        if hasattr(buffer, "itemsize")
                        else len(buffer)
                    )
                except (TypeError, AttributeError):
                    raise ValueError("Buffer must support the buffer protocol") from None

        status = clibmseed.mstl3_unpack_recordlist(
            self._parent_id,
            self._seg,
            buffer_ptr,
            buffer_size,
            verbose,
        )

        if status < 0:
            raise MiniSEEDError(status, "Error unpacking record list")
        else:
            return status

    def has_same_data(self, other: MS3TraceSeg) -> bool:
        """Compare trace segments for equivalent data

        Args:
            other: Another MS3TraceSeg to compare with

        Returns:
            True if segments have equivalent data, False otherwise
        """
        if not isinstance(other, MS3TraceSeg):
            return False

        return (
            self.starttime == other.starttime
            and self.endtime == other.endtime
            and self.samprate == other.samprate
            and self.samplecnt == other.samplecnt
            and self.datasamples == other.datasamples
            and self.datasize == other.datasize
            and self.numsamples == other.numsamples
            and self.sampletype == other.sampletype
        )


class MS3TraceID:
    """Wrapper around CFFI MS3TraceID structure

    This class supports list-like access to the trace segments:
    - len(traceid) returns the number of segments
    - traceid[i] returns the i-th segment
    - traceid[start:end] returns a slice of segments
    - for segment in traceid: iterates over all segments
    """

    def __init__(self, cffi_ptr: Any, parent_tracelist: Any = None) -> None:
        self._id = cffi_ptr
        self._parent_tracelist = parent_tracelist

    def __repr__(self) -> str:
        def indent_repr(thing):
            """Add two-space indentation to each line of repr(thing)"""
            return "\n".join("  " + line for line in repr(thing).split("\n"))

        # Create list of formatted strings
        if len(self) <= 5:
            formatted_lines = [indent_repr(traceid) for traceid in self]
        else:
            formatted_lines = [
                indent_repr(self[0]),
                indent_repr(self[1]),
                f"  ... {len(self) - 4} more",
                indent_repr(self[-2]),
                indent_repr(self[-1]),
            ]

        newline = "\n"
        return (
            f"MS3TraceID(sourceid: {self.sourceid}\n"
            f"         pubversion: {self.pubversion}\n"
            f"           earliest: {self.earliest_str(timeformat=TimeFormat.ISOMONTHDAY_DOY_Z)}\n"
            f"             latest: {self.latest_str(timeformat=TimeFormat.ISOMONTHDAY_DOY_Z)}\n"
            f"        numsegments: {len(self)}\n"
            f"{newline.join(formatted_lines)}"
            "\n)"
        )

    def __str__(self) -> str:
        return (
            f"{self.sourceid}, "
            f"v{self.pubversion}, "
            f"earliest: {self.earliest_str(timeformat=TimeFormat.ISOMONTHDAY_DOY_Z)}, "
            f"latest: {self.latest_str(timeformat=TimeFormat.ISOMONTHDAY_DOY_Z)}, "
            f"{len(self)} segments"
        )

    def __len__(self) -> int:
        """Return number of segments"""
        return self._id.numsegments

    def __iter__(self) -> Any:
        """Return iterator over segments"""
        current_segment = self._id.first
        while current_segment != ffi.NULL:
            yield MS3TraceSeg(current_segment, self._id, self._parent_tracelist)
            current_segment = current_segment.next

    def __getitem__(self, key: int | slice) -> Any:
        """Enable indexing and slicing access to segments"""
        if isinstance(key, slice):
            # Handle slice objects (e.g., traceid[1:3], traceid[::2])
            segment_list = list(self)
            return segment_list[key]
        elif isinstance(key, int):
            # Handle single integer index
            length = len(self)
            if length == 0:
                raise IndexError("list index out of range")

            # Handle negative indices
            if key < 0:
                key += length

            # Check bounds
            if key < 0 or key >= length:
                raise IndexError("list index out of range")

            # Find and return the segment at the specified index
            for i, segment in enumerate(self):
                if i == key:
                    return segment

            # This shouldn't happen if our logic is correct
            raise IndexError("list index out of range")
        else:
            raise TypeError("indices must be integers or slices")

    @property
    def sourceid(self) -> Optional[str]:
        """Return source ID as string"""
        result = cdata_to_string(self._id.sid)
        if result is None:
            return None
        return result

    @property
    def pubversion(self) -> int:
        """Return publication version"""
        return self._id.pubversion

    @property
    def earliest(self) -> int:
        """Return earliest time as nanoseconds since Unix/POSIX epoch"""
        return self._id.earliest

    @property
    def earliest_seconds(self) -> float:
        """Return earliest time as seconds since Unix/POSIX epoch"""
        return self._id.earliest / clibmseed.NSTMODULUS

    def earliest_str(
        self,
        timeformat: TimeFormat = TimeFormat.ISOMONTHDAY_Z,
        subsecond: SubSecond = SubSecond.NANO_MICRO_NONE,
    ) -> Optional[str]:
        """Return earliest time as formatted string"""
        result = nstime2timestr(self._id.earliest, timeformat, subsecond)
        if result is None:
            return None
        return result

    @property
    def latest(self) -> int:
        """Return latest time as nanoseconds since Unix/POSIX epoch"""
        return self._id.latest

    @property
    def latest_seconds(self) -> float:
        """Return latest time as seconds since Unix/POSIX epoch"""
        return self._id.latest / clibmseed.NSTMODULUS

    def latest_str(
        self,
        timeformat: TimeFormat = TimeFormat.ISOMONTHDAY_Z,
        subsecond: SubSecond = SubSecond.NANO_MICRO_NONE,
    ) -> Optional[str]:
        """Return latest time as formatted string"""
        result = nstime2timestr(self._id.latest, timeformat, subsecond)
        if result is None:
            return None
        return result


class MS3TraceList:
    """A container for a list of traces read from miniSEED

    If `file_name` is specified miniSEED will be read from the file.

    If `unpack_data` is True, the data samples will be decoded.

    If `skip_not_data` is True, bytes from the input stream will be skipped
    until a record is found.

    If `validate_crc` is True, the CRC will be validated if contained in
    the record (legacy miniSEED v2 contains no CRCs).  The CRC provides an
    internal integrity check of the record contents.

    The overall structure of the trace list list of trace IDs, each of which
    contains a list of trace segments illustrated as follows:
    - TraceList
      - TraceID
        - Trace Segment
        - Trace Segment
        - Trace Segment
        - ...
      - TraceID
        - Trace Segment
        - Trace Segment
        - ...
      - ...

    TraceIDs can be accessed via indexing and slicing:
    - `traces[0]` returns the first TraceID
    - `traces[1:3]` returns a slice of the TraceIDs
    - `for traceid in traces:` iterates over all TraceIDs

    Trace Segments can be accessed via indexing and slicing:
    - `traceid[0]` returns the first Trace Segment
    - `traceid[1:3]` returns a slice of the Trace Segments
    - `for segment in traceid:` iterates over all Trace Segments

    Example usage iterating over the trace list:
    ```
    >>> from pymseed import MS3TraceList

    >>> for traceid in MS3TraceList.from_file("examples/example_data.mseed"):
    ...     print(f"{traceid.sourceid}, {traceid.pubversion}")
    ...     for segment in traceid:
    ...         print(
    ...             f"  {segment.starttime_str()} - {segment.endtime_str()}, "
    ...             f"{segment.samprate} sps, {segment.samplecnt} samples"
    ...         )
    FDSN:IU_COLA_00_L_H_1, 4
      2010-02-27T06:50:00.069539Z - 2010-02-27T07:59:59.069538Z, 1.0 sps, 4200 samples
    FDSN:IU_COLA_00_L_H_2, 4
      2010-02-27T06:50:00.069539Z - 2010-02-27T07:59:59.069538Z, 1.0 sps, 4200 samples
    FDSN:IU_COLA_00_L_H_Z, 4
      2010-02-27T06:50:00.069539Z - 2010-02-27T07:59:59.069538Z, 1.0 sps, 4200 samples

    ```
    """

    def __init__(
        self,
        file_name=None,
        buffer=None,
        unpack_data=False,
        record_list=False,
        skip_not_data=False,
        validate_crc=True,
        split_version=False,
        verbose=0,
    ) -> None:
        # Initialize trace list - mstl3_init() returns an initialized pointer
        self._mstl = clibmseed.mstl3_init(ffi.NULL)

        if self._mstl == ffi.NULL:
            raise MiniSEEDError(clibmseed.MS_GENERROR, "Error initializing trace list")

        # Store filenames for record list functionality in C-compatible buffers
        self._c_file_names = []

        # Read specified file
        if file_name is not None:
            self.add_file(
                file_name,
                unpack_data,
                record_list,
                skip_not_data,
                validate_crc,
                split_version,
                verbose,
            )

        if buffer is not None:
            self.add_buffer(
                buffer,
                unpack_data,
                record_list,
                skip_not_data,
                validate_crc,
                split_version,
                verbose,
            )

    def __del__(self):
        """Destructor to ensure proper cleanup"""
        if self._mstl:
            mstl_ptr = ffi.new("MS3TraceList **")
            mstl_ptr[0] = self._mstl
            clibmseed.mstl3_free(mstl_ptr, 1)

    def __repr__(self) -> str:
        def indent_repr(thing):
            """Add two-space indentation to each line of repr(thing)"""
            return "\n".join("  " + line for line in repr(thing).split("\n"))

        # Create list of formatted strings
        if len(self) <= 5:
            formatted_lines = [indent_repr(traceid) for traceid in self]
        else:
            formatted_lines = [
                indent_repr(self[0]),
                indent_repr(self[1]),
                f"  ... {len(self) - 4} more",
                indent_repr(self[-2]),
                indent_repr(self[-1]),
            ]

        newline = "\n"
        return f"MS3TraceList(numtraceids: {len(self)}\n{newline.join(formatted_lines)}\n)"

    def __str__(self) -> str:
        def indent_str(thing):
            """Add two-space indentation to each line of str(thing)"""
            return "\n".join("  " + line for line in str(thing).split("\n"))

        # Create list of formatted strings
        if len(self) <= 5:
            formatted_lines = [indent_str(traceid) for traceid in self]
        else:
            formatted_lines = [
                indent_str(self[0]),
                indent_str(self[1]),
                f"  ... {len(self) - 4} more",
                indent_str(self[-2]),
                indent_str(self[-1]),
            ]

        newline = "\n"
        return f"Trace list with {len(self)} trace IDs\n{newline.join(formatted_lines)}\n"

    def __len__(self) -> int:
        """Return number of trace IDs in the list"""
        if self._mstl == ffi.NULL:
            return 0
        return int(self._mstl.numtraceids)

    def __iter__(self) -> Any:
        """Return iterator over trace IDs"""
        current_traceid = self._mstl.traces.next[0]
        while current_traceid != ffi.NULL:
            yield MS3TraceID(current_traceid, self)
            current_traceid = current_traceid.next[0]

    def __getitem__(self, key: int | slice) -> Any:
        """Enable indexing and slicing access to trace IDs"""
        if isinstance(key, slice):
            # Handle slice objects (e.g., traces[1:3], traces[::2])
            trace_list = list(self)
            return trace_list[key]
        elif isinstance(key, int):
            # Handle single integer index
            length = len(self)
            if length == 0:
                raise IndexError("list index out of range")

            # Handle negative indices
            if key < 0:
                key += length

            # Check bounds
            if key < 0 or key >= length:
                raise IndexError("list index out of range")

            # Find and return the trace ID at the specified index
            for i, traceid in enumerate(self):
                if i == key:
                    return traceid

            # This shouldn't happen if our logic is correct
            raise IndexError("list index out of range")
        else:
            raise TypeError("indices must be integers or slices")

    @property
    def numtraceids(self) -> int:
        """Return number of trace IDs in the list"""
        return len(self)

    def get_traceid(self, sourceid: str, version: int = 0) -> Optional[MS3TraceID]:
        """Get a specific trace ID from the list"""
        c_sourceid = ffi.new("char[]", sourceid.encode("utf-8"))

        traceid_ptr = clibmseed.mstl3_findID(self._mstl, c_sourceid, version, ffi.NULL)

        if traceid_ptr == ffi.NULL:
            return None

        return MS3TraceID(traceid_ptr, self)

    def sourceids(self) -> Any:
        """Return source IDs via a generator iterator"""
        for traceid in self:
            yield traceid.sourceid

    def print(
        self,
        details: int = 0,
        gaps: bool = False,
        versions: bool = False,
        timeformat: TimeFormat = TimeFormat.ISOMONTHDAY_Z,
    ) -> None:
        """Print trace list details"""
        clibmseed.mstl3_printtracelist(self._mstl, timeformat, details, gaps, versions)

    def add_file(
        self,
        file_name: str,
        unpack_data: bool = False,
        record_list: bool = False,
        skip_not_data: bool = False,
        validate_crc: bool = True,
        split_version: bool = False,
        verbose: int = 0,
    ) -> None:
        """Read miniSEED data from file and add to existing trace list

        This method reads miniSEED records from a file and adds the data they contain
        to the current trace list. Data are organized by source ID and time, with
        overlapping or adjacent data automatically merged into continuous segments.

        Args:
            file_name: Path to the miniSEED file to read
            unpack_data: If True, decode data samples immediately. If False, data
                samples remain packed and must be unpacked later with
                `unpack_recordlist()`. Default: False
            record_list: If True, maintain a list of original records for each
                trace segment. Required for `unpack_recordlist()` and allows
                access to individual record metadata. Default: False
                NOTE: the files must remain accessible to unpack data with a record list
            skip_not_data: If True, skip non-data records in the file instead
                of raising an error. Useful for files with mixed content. Default: False
            validate_crc: If True, validate CRC checksums if present in records
                (miniSEED v3 only). Provides integrity verification. Default: True
            split_version: If True, treat different publication versions as
                separate trace IDs. Default: False (merge by source ID only)
            verbose: Verbosity level for diagnostic output (0=quiet, 1-3=increasing
                detail). Default: 0

        Raises:
            MiniSEEDError: If file cannot be read or contains invalid data

        Note:
            This method adds data to the existing trace list. It does not replace
            existing data. To start fresh, create a new MS3TraceList instance.

        Examples:
            Basic usage examples:

            >>> from pymseed import MS3TraceList
            >>> traces = MS3TraceList()
            >>> traces.add_file("examples/example_data.mseed")
            >>> len(traces)
            3

            Unpacking data while reading for immediate access

            >>> traces = MS3TraceList()
            >>> traces.add_file("examples/example_data.mseed", unpack_data=True)
            >>> traces[0].sourceid
            'FDSN:IU_COLA_00_L_H_1'
            >>> segment1 = traces[0][0]
            >>> segment1.starttime_str()
            '2010-02-27T06:50:00.069539Z'
            >>> segment1.endtime_str()
            '2010-02-27T07:59:59.069538Z'
            >>> segment1.samprate
            1.0
            >>> segment1.samplecnt
            4200
            >>> segment1.numsamples
            4200
            >>> segment1.sampletype
            'i'

            Read with record list for later unpacking:

            >>> traces = MS3TraceList()
            >>> traces.add_file("examples/example_data.mseed", record_list=True)
            >>> total_samples = 0
            >>> for traceid in traces:
            ...     for segment in traceid:
            ...         # Unpack when desired, can unpacked to designated buffer
            ...         samples_count = segment.unpack_recordlist()
            ...         total_samples += samples_count
            >>> total_samples
            12600

            Add multiple files to same trace list, in this trivial example the
            same file is read twice, so the data are duplicated in the trace list:

            >>> traces = MS3TraceList()
            >>> traces.add_file("examples/example_data.mseed")
            >>> traces.add_file("examples/example_data.mseed")  # Appends to existing data
            >>> len(traces)
            3
            >>> traceid = traces[0]
            >>> traceid[0].samplecnt
            4200
            >>> len(traceid) # Two segments (duplicate in this case)
            2
            >>> # Compare using explicit data comparison (fast)
            >>> traceid[0].has_same_data(traceid[1])
            True

        """

        # Store files names for reference and use in record lists
        self._c_file_names.append(ffi.new("char[]", file_name.encode("utf-8")))
        c_file_name = self._c_file_names[-1]

        flags = 0
        if unpack_data:
            flags |= clibmseed.MSF_UNPACKDATA
        if record_list:
            flags |= clibmseed.MSF_RECORDLIST
        if skip_not_data:
            flags |= clibmseed.MSF_SKIPNOTDATA
        if validate_crc:
            flags |= clibmseed.MSF_VALIDATECRC

        # Create a reference to the current trace list pointer
        mstl_ptr = ffi.new("MS3TraceList **")
        mstl_ptr[0] = self._mstl

        status = clibmseed.ms3_readtracelist_selection(
            mstl_ptr,
            c_file_name,
            ffi.NULL,  # tolerance
            ffi.NULL,  # selections
            int(split_version),
            flags,
            verbose,
        )

        if status != clibmseed.MS_NOERROR:
            raise MiniSEEDError(status, f"Error reading file: {file_name}")

    def add_buffer(
        self,
        buffer: bytes,
        unpack_data: bool = False,
        record_list: bool = False,
        skip_not_data: bool = False,
        validate_crc: bool = True,
        split_version: bool = False,
        verbose: int = 0,
    ) -> None:
        """Read miniSEED data from a buffer and add to existing trace list

        This method reads miniSEED records from a bytes-like object and adds the
        data they contain to the current trace list. Data are organized by
        source ID and time, with overlapping or adjacent data automatically
        merged into continuous segments.

        Args:
            buffer: Bytes-like object containing miniSEED data
            unpack_data: If True, decode data samples immediately. If False, data
                samples remain packed and must be unpacked later with
                `unpack_recordlist()`. Default: False
            record_list: If True, maintain a list of original records for each
                trace segment. Required for `unpack_recordlist()` and allows
                access to individual record metadata. Default: False
                NOTE: the buffer must remain accessible to unpack data with a record list
                This can be quite tricky to achieve, for advanced use only
            skip_not_data: If True, skip non-data records in the buffer instead
                of raising an error. Useful for files with mixed content. Default: False
            validate_crc: If True, validate CRC checksums if present in records
                (miniSEED v3 only). Provides integrity verification. Default: True
            split_version: If True, treat different publication versions as
                separate trace IDs. Default: False (merge by source ID only)
            verbose: Verbosity level for diagnostic output (0=quiet, 1-3=increasing
                detail). Default: 0

        Raises:
            MiniSEEDError: If buffer cannot be read or contains invalid data

        Note:
            This method adds data to the existing trace list. It does not replace
            existing data. To start fresh, create a new MS3TraceList instance.
            To add data from a file, use `add_file()`.

        Examples:
            Read miniSEED data from a file into a buffer:

            >>> with open("examples/example_data.mseed", "rb") as f:
            ...     buffer = f.read()

            Basic usage examples:

            >>> from pymseed import MS3TraceList
            >>> traces = MS3TraceList()
            >>> traces.add_buffer(buffer)
            >>> len(traces)
            3

            Unpacking data while reading for immediate access

            >>> traces = MS3TraceList()
            >>> traces.add_buffer(buffer, unpack_data=True)
            >>> traces[0].sourceid
            'FDSN:IU_COLA_00_L_H_1'
            >>> segment1 = traces[0][0]
            >>> segment1.starttime_str()
            '2010-02-27T06:50:00.069539Z'
            >>> segment1.endtime_str()
            '2010-02-27T07:59:59.069538Z'
            >>> segment1.samprate
            1.0
            >>> segment1.samplecnt
            4200
            >>> segment1.numsamples
            4200
            >>> segment1.sampletype
            'i'

            Read with record list for later unpacking:

            >>> traces = MS3TraceList()
            >>> traces.add_buffer(buffer, record_list=True)
            >>> total_samples = 0
            >>> for traceid in traces:
            ...     for segment in traceid:
            ...         # Unpack when desired, can unpacked to designated buffer
            ...         samples_count = segment.unpack_recordlist()
            ...         total_samples += samples_count
            >>> total_samples
            12600

            Add multiple buffers to same trace list, in this trivial example the
            same file is read twice, so the data are duplicated in the trace list:

            >>> traces = MS3TraceList()
            >>> traces.add_buffer(buffer)
            >>> traces.add_buffer(buffer)  # Appends to existing data
            >>> len(traces)
            3
            >>> traceid = traces[0]
            >>> traceid[0].samplecnt
            4200
            >>> len(traceid) # Two segments (duplicate in this case)
            2

        """

        flags = 0
        if unpack_data:
            flags |= clibmseed.MSF_UNPACKDATA
        if record_list:
            flags |= clibmseed.MSF_RECORDLIST
        if skip_not_data:
            flags |= clibmseed.MSF_SKIPNOTDATA
        if validate_crc:
            flags |= clibmseed.MSF_VALIDATECRC

        # Create a reference to the current trace list pointer
        mstl_ptr = ffi.new("MS3TraceList **")
        mstl_ptr[0] = self._mstl

        # Validate that the buffer supports the buffer protocol
        try:
            buffer_ptr = ffi.from_buffer(buffer)
            buffer_length = len(buffer)
        except (TypeError, AttributeError):
            raise ValueError("Buffer must support the buffer protocol") from None

        status = clibmseed.mstl3_readbuffer_selection(
            mstl_ptr,
            buffer_ptr,
            buffer_length,
            int(split_version),
            flags,
            ffi.NULL,  # tolerance
            ffi.NULL,  # selections
            verbose,
        )

        if status < 0:
            raise MiniSEEDError(status, f"Error reading buffer (status: {status})")

    def add_data(
        self,
        sourceid: str,
        data_samples: Sequence[Any],
        sample_type: str,
        sample_rate: float,
        start_time_str: Optional[str] = None,
        start_time: Optional[int] = None,
        start_time_seconds: Optional[float] = None,
        publication_version: int = 0,
    ) -> None:
        """Add data samples to the trace list

        A segment of regularly sampled data values for the given source ID of
        the specific type and sample rate are added to the trace list.

        Args:
            sourceid: Source identifier for the trace (e.g., "FDSN:XX_STA__BHZ").
                Should follow FDSN Source Identifier format.
            data_samples: Sequence of data samples. Can be a Python list, numpy array,
                or any buffer-like object. Data type must match `sample_type`.
            sample_type: Data sample type code:
                - "i": 32-bit signed integers (int32)
                - "f": 32-bit floating point (float32)
                - "d": 64-bit floating point (float64)
                - "t": Text/character data (single bytes)
            sample_rate: Sample rate in samples per second (Hz) or period (seconds).
                Use positive values for samples/second, and negative values for sample period in seconds.
            start_time_str: Start time as formatted string (e.g., "2023-01-01T12:00:00.000Z").
                Mutually exclusive with start_time and start_time_seconds.
            start_time: Start time as nanoseconds since Unix epoch.
                Mutually exclusive with start_time_str and start_time_seconds.
            start_time_seconds: Start time as seconds since Unix epoch (float).
                Mutually exclusive with start_time_str and start_time.
            publication_version: Publication version number for the trace. Default: 0

        Raises:
            ValueError: If sample_type is invalid, time parameters are conflicting,
                or data_samples format is incompatible with sample_type
            MiniSEEDError: If the data cannot be added to the trace list

        Note:
            Data is automatically merged with existing segments based on source ID,
            time continuity, and sample rate compatibility. Adjacent or overlapping
            segments are combined when possible.

        Performance:
            The method attempts zero-copy optimization when data_samples is a compatible
            buffer (correct type and format). Otherwise, data is converted with a copy.
            Use arrays (or any buffer with memoryviews) with matching types for best performance.

        Examples:
            Basic usage with integer data:

            >>> from pymseed import MS3TraceList
            >>> traces = MS3TraceList()
            >>> data_series = [100, 105, 98, 102, 99, 103, 97]
            >>> traces.add_data(
            ...     sourceid="FDSN:XX_STA__BHZ",
            ...     data_samples=data_series,
            ...     sample_type="i",
            ...     sample_rate=20.0,
            ...     start_time_str="2023-01-01T00:00:00.000Z"
            ... )
            >>> len(traces)
            1
            >>> traces[0].sourceid
            'FDSN:XX_STA__BHZ'

            Multiple segments that get merged:

            >>> traces = MS3TraceList()
            >>> traces.add_data("FDSN:XX_STA__BH1", [1, 2, 3], "i", 10.0,
            ...                 start_time_str="2023-01-01T00:00:00.000Z")
            >>> traces.add_data("FDSN:XX_STA__BH1", [4, 5, 6], "i", 10.0,
            ...                 start_time_str="2023-01-01T00:00:00.300Z")
            >>> len(traces) # One traceID
            1
            >>> len(traces[0]) # One trace segment
            1
        """

        # Create an MS3Record to hold the data
        msr = MS3Record()
        msr.sourceid = sourceid
        msr.samprate = sample_rate
        msr.pubversion = publication_version

        # Set start time
        if start_time_str is not None:
            msr.set_starttime_str(start_time_str)
        elif start_time is not None:
            msr.starttime = start_time
        elif start_time_seconds is not None:
            msr.starttime_seconds = start_time_seconds
        else:
            raise ValueError(
                "Must specify one of start_time_str, start_time, or start_time_seconds"
            )

        # Request storing time of update in the trace list segment
        # This stores the update time as an nstime_t in the segment's private pointer (seg.prvtptr)
        flags = clibmseed.MSF_PPUPDATETIME

        # Set data samples array, type, and counts temporarily for potential zero-copy operations
        with msr.with_datasamples(data_samples, sample_type):
            # Add the MS3Record to the trace list, setting auto-heal flag to 1 (true)
            segptr = clibmseed.mstl3_addmsr_recordptr(
                self._mstl, msr._msr, ffi.NULL, 0, 1, flags, ffi.NULL
            )

        if segptr == ffi.NULL:
            raise MiniSEEDError(clibmseed.MS_GENERROR, "Error adding data samples")

    def _record_handler_wrapper(self, record: Any, record_length: int, handlerdata: Any) -> None:
        """Callback function for mstl3_pack()"""
        # Convert CFFI buffer to bytes for the handler
        record_bytes = ffi.buffer(record, record_length)[:]
        self._record_handler(record_bytes, self._record_handler_data)

    def pack(
        self,
        handler: Callable[[bytes, Any], None],
        handlerdata: Any = None,
        flush_data: bool = True,
        flush_idle_seconds: int = 0,
        record_length: int = 4096,
        encoding: DataEncoding = DataEncoding.STEIM1,
        format_version: Optional[int] = None,
        extra_headers: Optional[str] = None,
        verbose: int = 0,
    ) -> tuple[int, int]:
        """Pack trace list data into miniSEED records and call handler function for each record.

        This method packages the time series data from all traces in the trace list
        into miniSEED format records. For each generated record, the provided handler
        function is called with the record as a bytes object.

        Args:
            handler: Callback function that will be called for each packed record.
                Must accept two arguments: (record_bytes: bytes, userdata: Any).
                The record_bytes contains the complete miniSEED record.
            handlerdata: Optional user data passed to the handler function as the second argument.
                Can be any Python object (file handle, list, etc.).
            flush_data: If True, forces packing of all available data, even if it doesn't
                fill a complete record. If False, partial records at the end of traces
                may be held in internal buffers. Default is True.
            flush_idle_seconds: If > 0, forces flushing of data segments that have not been
                updated within the specified number of seconds. Default is 0 (disabled).
            record_length: Length of each miniSEED record in bytes. Must be a power of 2
                between 128 and 65536. Common values are 512, 4096, and 8192. Default is 4096.
            encoding: Data encoding format for compression. Options include:
                - DataEncoding.STEIM1: Steim-1 compression (default, good general purpose for 32-bit ints)
                - DataEncoding.STEIM2: Steim-2 compression
                - DataEncoding.INT16: 16-bit integers (no compression)
                - DataEncoding.INT32: 32-bit integers (no compression)
                - DataEncoding.FLOAT32: 32-bit IEEE floats
                - DataEncoding.FLOAT64: 64-bit IEEE doubles
                - DataEncoding.TEXT: Text encoding (UTF-8)
            format_version: miniSEED format version (2 or 3). If None, uses library default.
                Version 2 is legacy format, version 3 is latest standard.
            extra_headers: Optional extra header fields to include.
                Must be valid JSON string.
            verbose: Verbosity level for libmseed output (0=quiet, 1=info, 2=detailed).

        Returns:
            tuple[int, int]: A tuple containing:
                - packed_samples: Total number of data samples that were packed
                - packed_records: Total number of miniSEED records generated

        Raises:
            ValueError: If format_version is not 2 or 3, or if record_length is invalid.
            MiniSEEDError: If the underlying libmseed library encounters an error during packing.

        Examples:
            Simple example writing to a file:

            >>> # Create a trace list with some data
            >>> traces = MS3TraceList()
            >>> traces.add_data("FDSN:XX_STA__BHZ", [1, 2, 3, 4, 5], "i", 100.0, start_time_str="2023-01-01T00:00:00.000Z")

            >>> # Pack to file using a simple handler
            >>> def write_to_file(record_bytes, file_handle):
            ...     file_handle.write(record_bytes)

            >>> with open("output.mseed", "wb") as f: # doctest: +SKIP
            ...     packed_samples, packed_records = traces.pack(write_to_file, f)
            ...     print(f"Packed {packed_samples} samples into {packed_records} records")
            Packed 5 samples into 1 records

        Note:
            - The handler function is called once for each complete record generated
            - For large datasets, consider using streaming approaches with multiple pack() calls

        See also:
            - to_file()
        """

        # Set handler function as CFFI callback function
        self._record_handler = handler
        self._record_handler_data = handlerdata

        # Create callback function type and instance
        RECORD_HANDLER = ffi.callback("void(char *, int, void *)", self._record_handler_wrapper)

        pack_flags = 0
        if flush_data:
            pack_flags |= clibmseed.MSF_FLUSHDATA

        if format_version is not None:
            if format_version not in [2, 3]:
                raise ValueError(f"Invalid miniSEED format version: {format_version}")
            if format_version == 2:
                pack_flags |= clibmseed.MSF_PACKVER2

        packed_samples = ffi.new("int64_t *")

        c_extra = ffi.new("char[]", extra_headers.encode("utf-8")) if extra_headers else ffi.NULL

        packed_records = clibmseed.mstl3_pack(
            self._mstl,
            RECORD_HANDLER,
            ffi.NULL,
            record_length,
            encoding,
            packed_samples,
            pack_flags,
            verbose,
            c_extra,
        )

        if packed_records < 0:
            raise MiniSEEDError(packed_records, "Error packing miniSEED record(s)")

        # Iterate through all trace IDs and segments and flush idle data
        if flush_idle_seconds > 0:
            for traceid in self:
                for segment in traceid:
                    if segment.numsamples > 0 and segment._seg.prvtptr is not None:
                        # Pack the segment if it has been updated within the specified number of seconds

                        seg_packed_samples = ffi.new("int64_t *")
                        seg_packed_records = 0

                        # segment._seg.prvtptr is a void pointer to nstime_t value
                        nstime_ptr = ffi.cast("nstime_t *", segment._seg.prvtptr)
                        update_time_seconds = nstime_ptr[0] / clibmseed.NSTMODULUS

                        if (time() - update_time_seconds) > flush_idle_seconds:
                            seg_packed_records = clibmseed.mstl3_pack_segment(
                                self._mstl,
                                traceid._id,
                                segment._seg,
                                RECORD_HANDLER,
                                ffi.NULL,
                                record_length,
                                encoding,
                                seg_packed_samples,
                                pack_flags | clibmseed.MSF_FLUSHDATA,
                                verbose,
                                c_extra,
                            )

                            if seg_packed_records < 0:
                                raise MiniSEEDError(
                                    seg_packed_samples, "Error packing miniSEED records"
                                )

                            packed_samples[0] += seg_packed_samples[0]
                            packed_records += seg_packed_records

        return (packed_samples[0], packed_records)

    def to_file(
        self,
        filename: str,
        overwrite: bool = False,
        max_reclen: int = 4096,
        encoding: DataEncoding = DataEncoding.STEIM1,
        format_version: Optional[int] = None,
        verbose: int = 0,
    ) -> int:
        """Write trace list data to a miniSEED file.

        This method packages the time series data from all traces in the trace list
        into miniSEED format and writes them directly to a file. This is a convenience
        method that combines packing and file writing in a single operation.

        Args:
            filename: Path to the output miniSEED file. The file will be created if it
                doesn't exist. Directory must already exist.
            overwrite: If True, overwrites any existing file. If False and file exists,
                append data to the end of the file. Default is False for safety.
            max_reclen: Maximum record length in bytes. Must be a power of 2 between
                128 and 65536. Common values are 512, 4096, and 8192. Default is 4096.
            encoding: Data encoding format for compression. Options include:
                - DataEncoding.STEIM1: Steim-1 compression (default, good general purpose for 32-bit ints)
                - DataEncoding.STEIM2: Steim-2 compression
                - DataEncoding.INT16: 16-bit integers (no compression)
                - DataEncoding.INT32: 32-bit integers (no compression)
                - DataEncoding.FLOAT32: 32-bit IEEE floats
                - DataEncoding.FLOAT64: 64-bit IEEE doubles
                - DataEncoding.TEXT: Text encoding (UTF-8)
            format_version: miniSEED format version (2 or 3). If None, uses library default.
                Version 2 is legacy format, version 3 is latest standard.
            verbose: Verbosity level for libmseed output (0=quiet, 1=info, 2=detailed).

        Returns:
            int: Number of miniSEED records written to the file.

        Raises:
            ValueError: If format_version is not 2 or 3, or if max_reclen is invalid.
            MiniSEEDError: If the underlying libmseed library encounters an error during
                file writing (e.g., permission denied, disk full, invalid data).

        Examples:
            Simple file writing:

            >>> # Create a trace list with some data
            >>> traces = MS3TraceList()
            >>> traces.add_data("FDSN:XX_STA__BHZ", [1, 2, 3, 4, 5], "i", 100.0, start_time_str="2023-01-01T00:00:00Z")

            >>> # Write to file (basic usage)
            >>> records_written = traces.to_file("output.mseed") # doctest: +SKIP
            >>> print(f"Wrote {records_written} records to output.mseed") # doctest: +SKIP
            Wrote 1 records to output.mseed

            Writing with specific options:

            >>> # Write as miniSEED v2 with Steim-2 compression
            >>> records_written = traces.to_file( # doctest: +SKIP
            ...     "output.mseed",
            ...     overwrite=True,
            ...     format_version=2,
            ...     encoding=DataEncoding.STEIM2,
            ...     max_reclen=512
            ... )
            >>> print(f"Wrote {records_written} records to output.mseed") # doctest: +SKIP
            Wrote 1 records to output.mseed

        Note:
            - This method is more convenient than using pack() with a file handler
            - File permissions and disk space are checked during writing
            - All data is written in a single operation - use pack() for streaming
            - The output file contains all traces from the trace list
            - Record boundaries depend on max_reclen and data compression

        See also:
            - pack(): Lower-level method with custom record handlers
            - add_data(): Add time series data to the trace list
            - from_file(): Read miniSEED data from file
        """
        # Convert filename to bytes (C string)
        c_filename = ffi.new("char[]", filename.encode("utf-8"))

        pack_flags = 0
        if format_version is not None:
            if format_version not in [2, 3]:
                raise ValueError(f"Invalid miniSEED format version: {format_version}")
            if format_version == 2:
                pack_flags |= clibmseed.MSF_PACKVER2

        # Call the C function
        packed_records = clibmseed.mstl3_writemseed(
            self._mstl,
            c_filename,
            overwrite,
            max_reclen,
            encoding,
            pack_flags,
            verbose,
        )

        if packed_records < 0:
            raise MiniSEEDError(packed_records, "Error writing miniSEED file")

        return packed_records

    @classmethod
    def from_file(cls, filename, **kwargs):
        """Create MS3TraceList from a specified miniSEED file"""
        return cls(file_name=filename, **kwargs)

    @classmethod
    def from_buffer(cls, buffer: Any, **kwargs):
        """Create an MS3TraceList from miniSEED data in a memory buffer"""
        return cls(buffer=buffer, **kwargs)



from typing import Optional

from .util import error_string


class MiniSEEDError(ValueError):
    """Exception for libmseed return values"""

    def __init__(self, status_code: int, message: Optional[str] = None) -> None:
        self.status_code = status_code
        self.message = message

    def __str__(self) -> str:
        library_message = error_string(self.status_code)

        if library_message is None:
            library_message = f"Unknown error code: {self.status_code}"

        return f"{library_message} {':: ' + self.message if self.message else ''}"


class NoSuchSourceID(ValueError):
    """Exception for non-existent trace source IDs"""

    def __init__(self, sourceid: str) -> None:
        self.sourceid = sourceid

    def __str__(self) -> str:
        return f"Source ID not found: {self.sourceid}"



#!/usr/bin/env python3
"""
Generate miniSEED output from a continuous stream of data.

This program illustrates the use of an MS3TraceList to function as a temporary,
or transient, data buffer for a continuous stream of data, and using
MS3TraceList.pack() to continuously generate miniSEED output.

This pattern is useful for generating miniSEED output from a continuous stream
of data from any arbitrary source in a manner that creates full miniSEED records
as much as possible.  Data are either packed into records when a full record is
possible, when the data are idle for a specified number of seconds, or when the
program shuts down.

Usage: python continuous_miniseed_creation.py <output_file>

The output file will be continuously updated with new data.

For this example, a simple sine wave with is generated and used as the data
source.

Example usage:
  > python continuous_miniseed_creation.py output.mseed

This file is part of the pymseed package.
Copyright (c) 2025, EarthScope Data Services
"""

import argparse
import math
import signal
import threading
from collections.abc import Generator
from typing import Any

from pymseed import DataEncoding, MS3TraceList, NSTMODULUS, system_time

# Global flag and event for handling shutdown signals
shutdown_requested = False
shutdown_event = threading.Event()


def signal_handler(signum: int, frame: Any) -> None:
    """Handle shutdown signals (SIGTERM, SIGINT) by setting shutdown flag and event."""
    global shutdown_requested
    print(f"\nReceived signal {signum}, initiating shutdown...")
    shutdown_requested = True
    shutdown_event.set()  # Wake up any waiting threads immediately


def data_source(
    start_time: int = None,
    degree_offset: float = 0.0,
    sample_rate: float = 100.0,
    amplitude: float = 1.0,
    integer_samples: bool = False,
) -> Generator[tuple[list[float], int]]:
    """
    Generate a 1 Hertz sinusoidal signal for a time series starting at the
    specified start time and degree offset with the specified sample rate. The
    generator yields a list of new samples and the time of the first
    sample in the list (as a nanosecond timestamp).

    The number of samples returned is the number needed to fill the series since
    the last invocation.

    The start_time value is the time in nanosecond since the Unix epoch
    (nstime_t in libmseed), such as returned by pymseed.system_time() or
    pymseed.timestr2nstime(). If it is None, the current time is used.

    The degree_offset value is the phase offset in degrees, allowing for
    distinct sinusoids to be generated.

    The amplitude value is the maximum amplitude of the sinusoid. The sinusoid
    will range from -amplitude to +amplitude. Default is 1.0.

    The integer_samples value is a boolean indicating whether the samples should
    be integers or floats. Default is False (float samples).
    """
    if start_time is None:
        start_time = system_time()

    # Calculate the sample interval in nanoseconds
    sample_interval_ns = int(NSTMODULUS / sample_rate)

    # Convert degree offset to radians
    phase_offset = math.radians(degree_offset)

    # Keep track of the next sample time
    next_sample_time = start_time

    while True:
        current_time = system_time()

        # Generate samples until we catch up to current time
        samples = []
        first_sample_time = next_sample_time

        while next_sample_time <= current_time:
            # Calculate time for this sample in seconds
            sample_time_sec = (next_sample_time - start_time) / NSTMODULUS

            # Generate sinusoid sample
            sample_value = amplitude * math.sin(2 * math.pi * sample_time_sec + phase_offset)
            samples.append(int(sample_value) if integer_samples else sample_value)

            # Advance to next sample time
            next_sample_time += sample_interval_ns

        yield (samples, first_sample_time)


def create_continuous_miniseed(
    output_file: str, flush_idle_seconds: int, record_length: int, encoding: DataEncoding, verbose: int = 0
) -> None:
    """
    Create a miniSEED file from a continuous stream of data.

    Generates sinusoid data once per second for two source IDs until terminated
    by SIGTERM or Control-C.

    The data is written to the specified output file.
    """
    global shutdown_requested

    def file_handler(record_bytes: bytes, file_handle: Any) -> None:
        """Handler function to write packed miniSEED records to file."""
        file_handle.write(record_bytes)

    # Set up signal handlers for graceful shutdown
    signal.signal(signal.SIGTERM, signal_handler)
    signal.signal(signal.SIGINT, signal_handler)

    # Parameters for the data series
    sourceid_1 = "FDSN:XX_STA__H_X_1"
    sourceid_2 = "FDSN:XX_STA__B_X_2"
    sample_rate1 = 100.0
    sample_rate2 = 40.0

    # Determine sample type from encoding
    sample_type = "f" if encoding in [DataEncoding.FLOAT32, DataEncoding.FLOAT64] else "i"

    # Create data source generators for sinusoids with a 45 degree phase offset,
    # +/-100 amplitude range, and integer or float samples depending on the encoding
    sinusoid1 = data_source(
        degree_offset=0, sample_rate=sample_rate1, amplitude=100, integer_samples=sample_type == "i"
    )
    sinusoid2 = data_source(
        degree_offset=45, sample_rate=sample_rate2, amplitude=100, integer_samples=sample_type == "i"
    )

    # Create a trace buffer to function as a temporary data buffer
    # holding data until full-length records can be created or flushed
    # when idle.
    trace_buffer = MS3TraceList()

    # Print some information about the data series
    print(f"Starting continuous miniSEED creation to {output_file}")
    print(f"Flush idle time: {flush_idle_seconds} seconds")
    print(f"Record length: {record_length} bytes")
    print(f"Encoding: {encoding.name}")
    print(f"Verbosity level: {verbose}")
    print("Press Control-C to stop...\n")

    # Open the output file for writing
    with open(output_file, "wb") as output_handle:
        loop_count = 0
        while not shutdown_requested:
            try:
                # Get new data from generators, could be an arbitrary source
                new_data1, start_time1 = next(sinusoid1)

                # Only generate data for sourceid_2 every 5th loop (every 5 seconds)
                if loop_count % 5 == 0:
                    new_data2, start_time2 = next(sinusoid2)
                else:
                    new_data2 = []

                # Add any new data to trace buffer for sourceid_1
                if new_data1:
                    trace_buffer.add_data(
                        sourceid=sourceid_1,
                        data_samples=new_data1,
                        sample_type=sample_type,
                        sample_rate=sample_rate1,
                        start_time=start_time1,
                    )
                    if verbose > 1:
                        print(f"Added {len(new_data1)} samples to {sourceid_1}")

                # Add any new data to trace buffer for sourceid_2
                if new_data2:
                    trace_buffer.add_data(
                        sourceid=sourceid_2,
                        data_samples=new_data2,
                        sample_type=sample_type,
                        sample_rate=sample_rate2,
                        start_time=start_time2,
                    )
                    if verbose > 1:
                        print(f"Added {len(new_data2)} samples to {sourceid_2}")
                elif verbose > 2 and loop_count % 5 != 0:
                    print(f"Skipped data generation for {sourceid_2} (not 5th loop)")

                # Pack traces with idle data flushing, but don't flush all data yet
                packed_samples, packed_records = trace_buffer.pack(
                    handler=file_handler,
                    handlerdata=output_handle,
                    flush_data=False,
                    flush_idle_seconds=flush_idle_seconds,
                    record_length=record_length,
                    encoding=encoding,
                )

                if packed_records > 0 and verbose > 0:
                    print(f"Packed {packed_records} records with {packed_samples} samples")

                loop_count += 1

                # Wait for 1 second or until shutdown is requested (interruptible)
                # This allows the program to be terminated by SIGTERM/Control-C
                shutdown_event.wait(1)

            except KeyboardInterrupt:
                # Handle Control-C gracefully
                shutdown_requested = True
                shutdown_event.set()
                break

        # Perform final packing with flush_data=True to ensure all data is written
        print("\nPerforming final data creation to flush remaining data...")
        final_packed_samples, final_packed_records = trace_buffer.pack(
            handler=file_handler,
            handlerdata=output_handle,
            flush_data=True,
            record_length=record_length,
            encoding=encoding,
        )

        print(
            f"Final flush: packed {final_packed_records} records with {final_packed_samples} samples"
        )

        print(f"miniSEED creation complete. Output written to {output_file}")


def main():
    parser = argparse.ArgumentParser(
        description="Generate miniSEED output from a continuous stream of data until the program is terminated"
    )
    parser.add_argument("output_file", help="The file to write the miniSEED output to")
    parser.add_argument(
        "-f", "--flush_idle_seconds",
        type=int,
        default=10,
        help="The number of seconds of idle time before flushing the data",
    )
    parser.add_argument(
        "-r", "--record_length",
        type=int,
        default=512,
        help="Record length in bytes for miniSEED output (default: 512)",
    )
    parser.add_argument(
        "-e", "--encoding",
        type=str,
        choices=["STEIM1", "STEIM2", "FLOAT32", "FLOAT64", "INT16", "INT32"],
        default="STEIM1",
        help="Data encoding format for miniSEED output (default: STEIM1)",
    )
    parser.add_argument(
        "-v", "--verbose",
        action="count",
        default=0,
        help="Increase verbosity level (can be used multiple times: -v, -vv, -vvv)",
    )
    args = parser.parse_args()

    # Convert encoding string to DataEncoding enum
    encoding_map = {
        "STEIM1": DataEncoding.STEIM1,
        "STEIM2": DataEncoding.STEIM2,
        "FLOAT32": DataEncoding.FLOAT32,
        "FLOAT64": DataEncoding.FLOAT64,
        "INT16": DataEncoding.INT16,
        "INT32": DataEncoding.INT32,
    }
    encoding = encoding_map[args.encoding.upper()]

    # Create continuous miniSEED output until the program is terminated
    create_continuous_miniseed(
        args.output_file, args.flush_idle_seconds, args.record_length, encoding, args.verbose
    )


if __name__ == "__main__":
    main()



import time

import pytest

from pymseed import clibmseed
from pymseed.definitions import TimeFormat, SubSecond, DataEncoding
from pymseed.util import (
    nstime2timestr,
    timestr2nstime,
    sourceid2nslc,
    nslc2sourceid,
    encoding_string,
    error_string,
    sample_size,
    encoding_sizetype,
    sample_time,
    system_time,
)


class TestTimeConversion:
    """Test time conversion functions"""

    def test_nstime2timestr_basic(self):
        """Test basic nanosecond to time string conversion"""
        # Test with Unix epoch (0 nanoseconds = 1970-01-01T00:00:00Z)
        result = nstime2timestr(0)
        assert result == "1970-01-01T00:00:00Z"

        # Test with a known timestamp
        nstime = 1672574445123456789
        result = nstime2timestr(nstime)
        assert result == "2023-01-01T12:00:45.123456789Z"

    def test_nstime2timestr_different_formats(self):
        """Test different time format options"""
        nstime = 1672574445123456789

        # Test ISO month-day format (without Z)
        result = nstime2timestr(nstime, TimeFormat.ISOMONTHDAY)
        assert result == "2023-01-01T12:00:45.123456789"

        # Test SEED ordinal format
        result = nstime2timestr(nstime, TimeFormat.SEEDORDINAL)
        assert result == "2023,001,12:00:45.123456789"

        # Test space-separated format
        result = nstime2timestr(nstime, TimeFormat.ISOMONTHDAY_SPACE_Z)
        assert result == "2023-01-01 12:00:45.123456789Z"

    def test_nstime2timestr_different_subseconds(self):
        """Test different subsecond precision options"""
        nstime = 1672574445123456789  # With nanosecond precision

        # Test no subseconds
        result = nstime2timestr(nstime, subsecond=SubSecond.NONE)
        assert result == "2023-01-01T12:00:45Z"

        # Test microsecond precision
        result = nstime2timestr(nstime, subsecond=SubSecond.MICRO)
        assert result == "2023-01-01T12:00:45.123456Z"

        # Test nanosecond precision
        result = nstime2timestr(nstime, subsecond=SubSecond.NANO)
        assert result == "2023-01-01T12:00:45.123456789Z"

    def test_timestr2nstime_basic(self):
        """Test basic time string to nanosecond conversion"""
        # Test Unix epoch
        result = timestr2nstime("1970-01-01T00:00:00.000000000Z")
        assert result == 0

        # Test known timestamp
        result = timestr2nstime("2023-01-01T12:00:45.123456789Z")
        assert result == 1672574445123456789

    def test_timestr2nstime_various_formats(self):
        """Test time string parsing with various formats"""
        expected_nstime = 1672574445123456789

        # Test with Z suffix
        result = timestr2nstime("2023-01-01T12:00:45.123456789Z")
        assert result == expected_nstime

        # Test without Z suffix
        result = timestr2nstime("2023-01-01T12:00:45.123456789")
        assert result == expected_nstime

        # Test with space separator
        result = timestr2nstime("2023-01-01 12:00:45.123456789")
        assert result == expected_nstime

    def test_time_conversion_roundtrip(self):
        """Test that time conversion functions are inverse operations"""
        original_nstime = 1672574445123456789

        # Convert to string and back
        timestr = nstime2timestr(original_nstime)
        converted_nstime = timestr2nstime(timestr)

        assert converted_nstime == original_nstime

    def test_nstime2timestr_edge_cases(self):
        """Test edge cases for time conversion"""

        # Test with negative timestamp
        result = nstime2timestr(-1)
        assert result == "1969-12-31T23:59:59.999999999Z"


class TestSourceIdConversion:
    """Test source ID conversion functions"""

    def test_sourceid2nslc_basic(self):
        """Test basic source ID to NSLC conversion"""
        sourceid = "FDSN:XX_TEST__B_H_Z"
        net, sta, loc, chan = sourceid2nslc(sourceid)

        assert net == "XX"
        assert sta == "TEST"
        assert loc == ""
        assert chan == "BHZ"

    def test_sourceid2nslc_with_location(self):
        """Test source ID conversion with location code"""
        sourceid = "FDSN:IU_COLA_00_B_H_Z"
        net, sta, loc, chan = sourceid2nslc(sourceid)

        assert net == "IU"
        assert sta == "COLA"
        assert loc == "00"
        assert chan == "BHZ"

    def test_nslc2sourceid_basic(self):
        """Test basic NSLC to source ID conversion"""
        result = nslc2sourceid("IU", "ANMO", "00", "BHZ")
        assert result == "FDSN:IU_ANMO_00_B_H_Z"

    def test_nslc2sourceid_empty_location(self):
        """Test NSLC to source ID conversion with empty location"""
        result = nslc2sourceid("IU", "ANMO", "", "BHZ")
        assert result == "FDSN:IU_ANMO__B_H_Z"

    def test_sourceid_conversion_roundtrip(self):
        """Test that source ID conversion functions are inverse operations"""
        original_net, original_sta, original_loc, original_chan = "IU", "ANMO", "00", "BHZ"

        # Convert to source ID and back
        sourceid = nslc2sourceid(original_net, original_sta, original_loc, original_chan)
        net, sta, loc, chan = sourceid2nslc(sourceid)

        assert net == original_net
        assert sta == original_sta
        assert loc == original_loc
        assert chan == original_chan


class TestEncodingFunctions:
    """Test encoding-related utility functions"""

    def test_encoding_string_basic(self):
        """Test getting encoding description strings"""
        # Test known encoding types using integer values
        result = encoding_string(int(DataEncoding.TEXT))
        assert "text" in result.lower() or "utf" in result.lower()

        result = encoding_string(int(DataEncoding.INT32))
        assert "32" in result and ("int" in result.lower() or "integer" in result.lower())

        result = encoding_string(int(DataEncoding.FLOAT32))
        assert "32" in result and "float" in result.lower()

        result = encoding_string(int(DataEncoding.FLOAT64))
        assert "64" in result and "float" in result.lower()

        result = encoding_string(int(DataEncoding.STEIM1))
        assert "steim" in result.lower() and "1" in result

        result = encoding_string(int(DataEncoding.STEIM2))
        assert "steim" in result.lower() and "2" in result

    def test_sample_size_basic(self):
        """Test getting sample sizes for different encodings"""
        # ms_samplesize expects a char sampletype, not encoding number
        # Test with sample type characters as bytes
        assert sample_size(b"i") == 4  # integer type
        assert sample_size(b"f") == 4  # float type
        assert sample_size(b"d") == 8  # double type
        assert sample_size(b"t") == 1  # text type

    def test_encoding_sizetype_basic(self):
        """Test getting encoding size and type information"""
        # Test INT16
        size, sample_type = encoding_sizetype(int(DataEncoding.INT16))
        assert size == 4  # From the C code, INT16 also returns 4
        assert sample_type == "i"  # integer type

        # Test INT32
        size, sample_type = encoding_sizetype(int(DataEncoding.INT32))
        assert size == 4
        assert sample_type == "i"  # integer type

        # Test FLOAT32
        size, sample_type = encoding_sizetype(int(DataEncoding.FLOAT32))
        assert size == 4
        assert sample_type == "f"  # float type

        # Test FLOAT64
        size, sample_type = encoding_sizetype(int(DataEncoding.FLOAT64))
        assert size == 8
        assert sample_type == "d"  # double type (not 'f')

        # Test Steim1
        size, sample_type = encoding_sizetype(int(DataEncoding.STEIM1))
        assert size == 4
        assert sample_type == "i"  # integer type

        # Test Steim2
        size, sample_type = encoding_sizetype(int(DataEncoding.STEIM2))
        assert size == 4
        assert sample_type == "i"  # integer type

    def test_encoding_sizetype_invalid_encoding(self):
        """Test error handling for invalid encoding"""
        with pytest.raises(ValueError):
            encoding_sizetype(255)  # Invalid encoding number that fits in uint8_t


class TestErrorHandling:
    """Test error handling functions"""

    def test_error_string_basic(self):
        """Test getting error description strings"""
        # Test with error code 0 (usually means no error/success)
        result = error_string(0)
        assert isinstance(result, str)
        assert len(result) > 0

        # Test with common error codes
        for error_code in [-1, -2, -3]:
            result = error_string(error_code)
            assert isinstance(result, str)
            assert len(result) > 0


class TestSampleTime:
    """Test sample time calculation function"""

    def test_sample_time_basic(self):
        """Test basic sample time calculations"""
        # Test with 1 Hz sampling rate
        base_time = 1672574445000000000  # Some base time in nanoseconds
        samprate = 1.0  # 1 Hz = 1 sample per second

        # Sample at offset 0 should be at base time
        result = sample_time(base_time, 0, samprate)
        assert result == base_time

        # Sample at offset 1 should be 1 second later (1e9 nanoseconds)
        result = sample_time(base_time, 1, samprate)
        expected = base_time + int(1e9)  # 1 second in nanoseconds
        assert result == expected

    def test_sample_time_high_frequency(self):
        """Test sample time calculations with high frequency sampling"""
        base_time = 1672574445000000000
        samprate = 100.0  # 100 Hz = 100 samples per second

        # Sample at offset 0
        result = sample_time(base_time, 0, samprate)
        assert result == base_time

        # Sample at offset 1 should be 0.01 seconds later (1e7 nanoseconds)
        result = sample_time(base_time, 1, samprate)
        expected = base_time + int(1e7)  # 0.01 seconds in nanoseconds
        assert result == expected

        # Sample at offset 100 should be 1 second later
        result = sample_time(base_time, 100, samprate)
        expected = base_time + int(1e9)  # 1 second in nanoseconds
        assert result == expected

    def test_sample_time_fractional_offset(self):
        """Test sample time calculations with fractional sampling rates"""
        base_time = 1672574445000000000
        samprate = 0.1  # 0.1 Hz = 1 sample per 10 seconds

        # Sample at offset 1 should be 10 seconds later
        result = sample_time(base_time, 1, samprate)
        expected = base_time + int(10e9)  # 10 seconds in nanoseconds
        assert result == expected

    def test_sample_time_zero_offset(self):
        """Test sample time calculation with zero offset"""
        base_time = 1672574445123456789
        result = sample_time(base_time, 0, 50.0)
        assert result == base_time


class TestSystemtime:
    """Test systemtime function"""

    def test_systemtime(self):
        """Test systemtime function"""
        system_time_ns = system_time()

        system_time_sec = float(system_time_ns / clibmseed.NSTMODULUS)

        # Verify within a 100 milliseconds of time.time()
        assert abs(time.time() - system_time_sec) < 0.1, (
            f"Time difference {abs(time.time() - system_time_sec) * 1e3:.3f} ms exceeds 100 ms tolerance"
        )



import os

import pytest

from pymseed import MS3Record, DataEncoding, TimeFormat, SubSecond

test_dir = os.path.abspath(os.path.dirname(__file__))
test_path3 = os.path.join(test_dir, "data", "testdata-COLA-signal.mseed3")
test_path2 = os.path.join(test_dir, "data", "testdata-COLA-signal.mseed2")


def test_msrecord_read_buffer_details():
    # Read data from test file into a buffer
    with open(test_path3, "rb") as fp:
        buffer = bytearray(fp.read())

        with MS3Record.from_buffer(buffer, unpack_data=True) as msreader:

            # Read first record
            msr = msreader.read()

            assert msr.reclen == 542
            assert msr.swapflag == 2
            assert msr.swapflag_dict() == {
                "header_swapped": False,
                "payload_swapped": True,
            }
            assert msr.sourceid == "FDSN:IU_COLA_00_B_H_1"
            assert msr.formatversion == 3
            assert msr.flags == 4
            assert msr.flags_dict() == {"clock_locked": True}
            assert msr.starttime == 1267253400019539000
            assert msr.starttime_seconds == 1267253400.019539
            assert (
                msr.starttime_str(timeformat=TimeFormat.ISOMONTHDAY_Z)
                == "2010-02-27T06:50:00.019539Z"
            )
            assert (
                msr.starttime_str(
                    timeformat=TimeFormat.SEEDORDINAL, subsecond=SubSecond.NONE
                )
                == "2010,058,06:50:00"
            )
            assert msr.samprate == 20.0
            assert msr.samprate_raw == 20.0
            assert msr.encoding == DataEncoding.STEIM2
            assert msr.encoding_str() == "STEIM-2 integer compression"
            assert msr.pubversion == 4
            assert msr.samplecnt == 296
            assert msr.crc == 1977151071
            assert msr.extralength == 33
            assert msr.datalength == 448
            assert msr.extra == '{"FDSN":{"Time":{"Quality":100}}}'
            assert msr.numsamples == 296
            assert msr.sampletype == "i"
            assert msr.endtime == 1267253414769539000
            assert msr.endtime_seconds == 1267253414.769539

            # Data sample array tests
            data = msr.datasamples

            # Check first 6 samples
            assert data[0:6].tolist() == [-502916, -502808, -502691, -502567, -502433, -502331]

            # Check last 6 samples
            assert data[-6:].tolist() == [-508722, -508764, -508809, -508866, -508927, -508986]


def test_msrecord_numpy():
    np = pytest.importorskip("numpy")

    # Read data from test file into a buffer
    with open(test_path3, "rb") as fp:
        buffer = bytearray(fp.read())

        with MS3Record.from_buffer(buffer, unpack_data=True) as msreader:

            # Read first record
            msr = msreader.read()

            # Data sample array tests
            data = msr.np_datasamples

            # Check first 6 samples
            assert np.all(
                data[0:6].tolist() == [-502916, -502808, -502691, -502567, -502433, -502331]
            )

            # Check last 6 samples
            assert np.all(
                data[-6:].tolist() == [-508722, -508764, -508809, -508866, -508927, -508986]
            )


def test_msrecord_read_buffer_summary():
    # Read data from test file into a buffer
    with open(test_path2, "rb") as fp:
        buffer = bytearray(fp.read())

        record_count = 0
        sample_count = 0

        # Direct iteration without context manager
        for msr in MS3Record.from_buffer(buffer):
            record_count += 1
            sample_count += msr.samplecnt

        assert record_count == 1141
        assert sample_count == 252000



#!/usr/bin/env python3
"""
Read miniSEED file(s) from a stream (stdin), select those that
fall within the selected earliest and latest times, and write out
to a stream (stdout). Records that contain the selected times are
trimmed to the selected times.

Example usage:
 > stream_timewindow.py --earliest 2010-02-27T07:00:00 --latest 2010-02-27T07:10:00 < example_data.mseed > windowed.mseed

This file is part of the pymseed package.
Copyright (c) 2025, EarthScope Data Services
"""

import argparse
import sys

from pymseed import MS3Record, timestr2nstime, NSTMODULUS

def process_stream(args):
    """Process miniSEED records from stdin, applying time window selection."""
    records_written = 0
    bytes_written = 0

    print("Reading miniSEED from stdin, writing to stdout", file=sys.stderr)

    # Read miniSEED from stdin
    with MS3Record.from_file(sys.stdin.fileno()) as msreader:
        for record in msreader:
            # Skip records completely outside the time window
            if _record_outside_timewindow(record, args.earliest, args.latest):
                continue

            # Check if record needs trimming
            output_record = record.record
            if _record_needs_trimming(record, args.earliest, args.latest):
                trimmed_record = trim_record(record, args.earliest, args.latest)
                if trimmed_record:
                    output_record = trimmed_record

            # Write record to stdout
            sys.stdout.buffer.write(output_record)
            records_written += 1
            bytes_written += record.reclen

    print(f"Wrote {records_written} records, {bytes_written} bytes", file=sys.stderr)


def _record_outside_timewindow(record, earliest, latest):
    """Check if a record is completely outside the specified time window."""
    # Record ends before earliest time
    if earliest and record.endtime < earliest:
        return True

    # Record starts after latest time
    if latest and record.starttime > latest:
        return True

    return False


def _record_needs_trimming(record, earliest, latest):
    """Check if a record overlaps with time window boundaries and needs trimming."""
    needs_early_trim = earliest and record.starttime < earliest <= record.endtime
    needs_late_trim = latest and record.starttime <= latest < record.endtime
    return needs_early_trim or needs_late_trim


def trim_record(record, earliest, latest):
    """Trim a miniSEED record to the specified start and end times."""
    # Cannot trim time coverage of a record with no coverage
    if record.samplecnt == 0 and record.samprate == 0.0:
        return None

    # Re-parse the record and decode the data samples
    buffer = bytearray(record.record)  # Mutable/writable buffer required
    with MS3Record.from_buffer(buffer, unpack_data=True) as msreader:
        record = msreader.read()

        trimmed_data = _trim_data_samples(record, earliest, latest)
        if not trimmed_data:
            return None

        # Pack the trimmed record
        return _pack_trimmed_record(record, trimmed_data)


def _trim_data_samples(record, earliest, latest):
    """Extract and trim data samples based on time window."""
    data_samples = record.datasamples[:]
    start_time = record.starttime
    end_time = record.endtime
    sample_period_ns = int(NSTMODULUS / record.samprate)

    # Trim early samples to the earliest time
    if earliest and start_time < earliest <= end_time:
        count = 0
        while start_time < earliest:
            start_time += sample_period_ns
            count += 1
        data_samples = data_samples[count:]

    # Trim late samples to the latest time
    if latest and start_time <= latest < end_time:
        count = 0
        while end_time > latest:
            end_time -= sample_period_ns
            count += 1
        data_samples = data_samples[:-count] if count > 0 else data_samples

    return {
        'data': data_samples,
        'start_time': start_time
    }


def _pack_trimmed_record(record, trimmed_data):
    """Pack a record with trimmed data samples."""
    # Use a simple approach to capture the packed record
    packed_record = None

    def record_handler(record_bytes, handler_data):
        nonlocal packed_record
        packed_record = bytes(record_bytes)

    # Update the record's start time and pack with trimmed data
    record.starttime = trimmed_data['start_time']
    record.pack(record_handler, data_samples=trimmed_data['data'], sample_type=record.sampletype)

    return packed_record


def parse_timestr(timestr):
    """
    Helper for argparse to convert a time string to a nanosecond time value.
    """
    try:
        return timestr2nstime(timestr)
    except ValueError:
        raise argparse.ArgumentTypeError(f"Invalid time string: {timestr}") from None


def main():
    """Main entry point for the script."""
    parser = argparse.ArgumentParser(
        description="Stream miniSEED records with time window selection",
        epilog="Reads from stdin and writes to stdout. Records overlapping the "
               "time window boundaries are trimmed to fit within the window."
    )
    parser.add_argument(
        "--earliest", "-e",
        type=parse_timestr,
        help="Earliest time to include (ISO format: YYYY-MM-DDTHH:MM:SS)"
    )
    parser.add_argument(
        "--latest", "-l",
        type=parse_timestr,
        help="Latest time to include (ISO format: YYYY-MM-DDTHH:MM:SS)"
    )

    args = parser.parse_args()

    # Validate time arguments
    if args.earliest and args.latest and args.earliest > args.latest:
        parser.error("Earliest time cannot be after latest time")

    if not args.earliest and not args.latest:
        parser.error("At least one of --earliest or --latest must be specified")

    try:
        process_stream(args)
    except BrokenPipeError:
        # Handle broken pipe gracefully (e.g., when piping to head)
        pass
    except KeyboardInterrupt:
        print("\nInterrupted by user", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import importlib.util
import sys
from pathlib import Path
from setuptools import setup
from cffi import FFI

# --- Configuration ---
# Get the absolute path to the directory containing setup.py
here = Path(__file__).parent.resolve()
cffi_module_path = here / "cffi_module"
libmseed_path = cffi_module_path / "libmseed"

# --- CFFI Setup ---
# Import the CFFI definitions from the separate definitions file
spec = importlib.util.spec_from_file_location(
    "cffi_defs", cffi_module_path / "cffi_defs.py"
)
cffi_defs = importlib.util.module_from_spec(spec)
spec.loader.exec_module(cffi_defs)

# Create FFI instance and configure it with the C definitions
ffi = FFI()
ffi.cdef(cffi_defs.LIBRARY_CDEF)

# Find all C source files.
#
# IMPORTANT: Paths must be explicitly relative to the 'setup.py' directory
# for setuptools to work correctly, especially with 'python -m build'.
# We use pathlib to find the absolute paths and then make them relative to 'here'.
c_sources = [str(p.relative_to(here)) for p in libmseed_path.glob("*.c")]

# --- Platform-specific compiler options ---
if sys.platform.startswith("win"):
    # Windows-specific options
    extra_compile_args = ["/O2"]
    extra_link_args = []
    define_macros = [("_CRT_SECURE_NO_WARNINGS", None)]
else:
    # Unix-like systems (Linux, macOS)
    extra_compile_args = ["-O2"]
    extra_link_args = []
    define_macros = []

# --- CFFI Source Configuration ---
# Configure the CFFI extension module
ffi.set_source(
    "_libmseed_cffi",
    # This is the C code that will be compiled. It includes the header
    # which makes all the C functions available to the CFFI module.
    '#include "libmseed.h"',
    # Provide the list of all C source files to be compiled together.
    sources=c_sources,
    # Provide the include directory, also as a relative path.
    include_dirs=[str(libmseed_path.relative_to(here))],
    # Pass platform-specific compiler and linker arguments.
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
    define_macros=define_macros,
)

# --- Setuptools Configuration ---
setup(
    # Tell setuptools that CFFI is a build-time requirement.
    setup_requires=["cffi>=1.0.0"],
    # And also an install-time requirement for the end-user.
    install_requires=["cffi>=1.0.0"],
    # This is where the magic happens: ffi.distutils_extension() creates
    # the Extension object that setuptools will build.
    ext_modules=[ffi.distutils_extension()],
    # The compiled extension will be placed inside the 'pymseed' package.
    ext_package="pymseed",
    # Wheels with compiled extensions are not zip-safe.
    zip_safe=False,
)



import pytest
import sys
import os
from pymseed import MS3Record, DataEncoding, TimeFormat, SubSecond
from pymseed.exceptions import MiniSEEDError

test_dir = os.path.abspath(os.path.dirname(__file__))
test_path3 = os.path.join(test_dir, "data", "testdata-COLA-signal.mseed3")
test_path2 = os.path.join(test_dir, "data", "testdata-COLA-signal.mseed2")


def test_msrecord_read_record_details():
    with MS3Record.from_file(test_path3, unpack_data=True) as msreader:

        # Read first record
        msr = msreader.read()

        assert msr.reclen == 542
        assert msr.swapflag == 2
        assert msr.swapflag_dict() == {"header_swapped": False, "payload_swapped": True}
        assert msr.sourceid == "FDSN:IU_COLA_00_B_H_1"
        assert msr.formatversion == 3
        assert msr.flags == 4
        assert msr.flags_dict() == {"clock_locked": True}
        assert msr.starttime == 1267253400019539000
        assert msr.starttime_seconds == 1267253400.019539
        assert (
            msr.starttime_str(timeformat=TimeFormat.ISOMONTHDAY_Z)
            == "2010-02-27T06:50:00.019539Z"
        )
        assert (
            msr.starttime_str(
                timeformat=TimeFormat.SEEDORDINAL, subsecond=SubSecond.NONE
            )
            == "2010,058,06:50:00"
        )
        assert msr.samprate == 20.0
        assert msr.samprate_raw == 20.0
        assert msr.encoding == DataEncoding.STEIM2
        assert msr.encoding_str() == "STEIM-2 integer compression"
        assert msr.pubversion == 4
        assert msr.samplecnt == 296
        assert msr.crc == 1977151071
        assert msr.extralength == 33
        assert msr.datalength == 448
        assert msr.extra == '{"FDSN":{"Time":{"Quality":100}}}'
        assert msr.numsamples == 296
        assert msr.sampletype == "i"
        assert msr.endtime == 1267253414769539000
        assert msr.endtime_seconds == 1267253414.769539

        # Check first 6 samples
        assert msr.datasamples[0:6].tolist() == [-502916, -502808, -502691, -502567, -502433, -502331]

        # Check last 6 samples
        assert msr.datasamples[-6:].tolist() == [-508722, -508764, -508809, -508866, -508927, -508986]


def test_msrecord_read_record_details_fd():
    # Test reading from a file descriptor - we simulate this using the buffer reader

    # File descriptor support is not implemented on Windows
    if sys.platform.lower().startswith("win"):
        return

    # Using a file for testing, but this could be stdin or any other input stream
    file_descriptor = None
    with open(test_path2, "rb", buffering=0) as fp:
        original_file_descriptor = fp.fileno()
        file_descriptor = os.dup(original_file_descriptor)

    # Provide the reader with the file descriptor
    with MS3Record.from_file(file_descriptor, unpack_data=True) as msreader:
        # Read first record
        msr = msreader.read()

        # Verify we got a valid record
        assert msr is not None

        # Check first 6 samples
        assert msr.datasamples[0:6].tolist() == [-502916, -502808, -502691, -502567, -502433, -502331]

        # Check last 6 samples
        assert msr.datasamples[-6:].tolist() == [-508722, -508764, -508809, -508866, -508927, -508986]


def test_msrecord_read_records_summary():
    record_count = 0
    sample_count = 0

    # Direct iteration without context manager
    for msr in MS3Record.from_file(test_path2):
        record_count += 1
        sample_count += msr.samplecnt

    assert record_count == 1141
    assert sample_count == 252000


def test_msrecord_nosuchfile():
    with pytest.raises(MiniSEEDError):
        with MS3Record.from_file("NOSUCHFILE") as msreader:
            msr = msreader.read()



"""
Core library interface for pymseed using CFFI

"""

from typing import Any, Optional

try:
    # This is the correct pattern: import the ffi and lib objects
    # directly FROM the compiled _libmseed_cffi module.
    from ._libmseed_cffi import ffi
    from ._libmseed_cffi import lib as clibmseed  # noqa: F401

except ImportError as exc:
    # The friendly error message is still a good idea.
    # The 'from exc' part preserves the original traceback for debugging.
    raise ImportError(
        "Could not import the CFFI-based C extension module.\n"
        "This is likely because the package is not installed correctly.\n"
        "Please make sure the package is installed, for example by running:\n"
        "  pip install .\n"
        "or for development:\n"
        "  pip install -e ."
    ) from exc


def cdata_to_string(cdata: Any, encoding: str = "utf-8") -> Optional[str]:
    """
    Convert C string to Python string.  If the C string is NULL, return None.

    Args:
        cdata: CFFI cdata char*
        encoding: String encoding

    Returns:
        Python string
    """
    if cdata == ffi.NULL:
        return None
    else:
        return str(ffi.string(cdata).decode(encoding))



#!/usr/bin/env python3
"""
Read miniSEED file(s) from a stream, accumulate stats, and write to a stream.

For this illustration input is stdin, output is stdout, and stats are printed
to stderr on completion.

Example usage:
  cat example_data.mseed | stream_stats.py > output.mseed

This file is part of the pymseed package.
Copyright (c) 2025, EarthScope Data Services
"""

import pprint
import sys

from pymseed import MS3Record, nstime2timestr


def create_initial_stats():
    """Create a new stats dictionary with default values."""
    return {
        "record_count": 0,
        "sample_count": 0,
        "bytes": 0,
        "pubversions": [],
        "formatversions": [],
        "earliest": None,
        "latest": None,
    }


def update_stats(stats, record):
    """Update statistics with data from a miniSEED record."""
    # Update counters
    stats["record_count"] += 1
    stats["sample_count"] += record.samplecnt
    stats["bytes"] += record.reclen

    # Track unique publication versions
    if record.pubversion not in stats["pubversions"]:
        stats["pubversions"].append(record.pubversion)

    # Track unique format versions
    if record.formatversion not in stats["formatversions"]:
        stats["formatversions"].append(record.formatversion)

    # Track earliest sample time (fix: should use < for earliest)
    if stats["earliest"] is None or record.starttime < stats["earliest"]:
        stats["earliest"] = record.starttime

    # Track latest sample time (fix: should use > for latest)
    if stats["latest"] is None or record.endtime > stats["latest"]:
        stats["latest"] = record.endtime


def main():
    """Main processing function."""
    trace_stats = {}

    print("Reading miniSEED from stdin, writing to stdout", file=sys.stderr)

    # Read miniSEED from stdin and process each record
    with MS3Record.from_file(sys.stdin.fileno()) as reader:
        for record in reader:
            # Get or create stats for this source ID
            if record.sourceid not in trace_stats:
                trace_stats[record.sourceid] = create_initial_stats()

            # Update statistics for this trace
            update_stats(trace_stats[record.sourceid], record)

            # Write raw miniSEED record to stdout
            sys.stdout.buffer.write(record.record)

    # Add human-readable time strings
    for stats in trace_stats.values():
        if stats["earliest"] is not None:
            stats["earliest_str"] = nstime2timestr(stats["earliest"])
        if stats["latest"] is not None:
            stats["latest_str"] = nstime2timestr(stats["latest"])

    # Print statistics to stderr
    printer = pprint.PrettyPrinter(stream=sys.stderr, indent=4, sort_dicts=False)
    printer.pprint(trace_stats)


if __name__ == "__main__":
    main()
